"""Models and functions used for representing states."""

from abc import ABC
from dataclasses import asdict, dataclass
from typing import Any, Dict


@dataclass
class State(ABC):
    """A state."""

    def as_dict(self) -> Dict[str, Any]:
        """
        Get the fields a a new dict.

        Returns:
            a dict containing the fields values.
        """
        return asdict(self)



"""Models used for the configuration of YesWeHack."""
from typing import Any, Dict, List, Optional, Union, cast

from ywh2bt.core.configuration.attribute import (
    Attribute,
    AttributesContainer,
    AttributesContainerDict,
    AttributesContainerList,
    BoolAttributeType,
    ExportableList,
    StrAttributeType,
)
from ywh2bt.core.configuration.headers import Headers
from ywh2bt.core.configuration.validator import (
    dict_has_non_blank_key_validator,
    length_one_validator,
    not_blank_validator,
    not_empty_validator,
    url_validator,
)


class OAuthSettings(AttributesContainer):
    """OAuth settings."""

    client_id: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='Client ID',
        description='OAuthSettings client ID',
        required=True,
        validator=not_blank_validator,
    )
    client_secret: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='Secret',
        description='OAuthSettings client secret',
        required=True,
        secret=True,
        validator=not_blank_validator,
    )
    redirect_uri: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='Redirect URI',
        description='OAuthSettings redirect URI',
        required=True,
        validator=url_validator,
    )

    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        redirect_uri: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the settings.

        Args:
            client_id: an OAuth v2 client ID
            client_secret: an OAuth v2 client secret
            redirect_uri: a redirect URI
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri

    def validate(
        self,
    ) -> None:
        """Validate the object."""
        attributes = (
            self.client_id,
            self.client_secret,
            self.redirect_uri,
        )
        if not any(attributes):
            return
        super().validate()


class Bugtrackers(ExportableList[str, str]):
    """A list of bugtrackers."""


BugtrackersNameAttributeType = Union[
    Optional[Bugtrackers],
    Attribute[Bugtrackers],
]
BugtrackersNameInitType = Union[
    Optional[Bugtrackers],
    Optional[List[str]],
]


class SynchronizeOptions(AttributesContainer):
    """Synchronization option."""

    upload_private_comments: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description='Upload private comments',
        description='Upload the report private comments into the bug trackers',
        default=False,
    )
    upload_public_comments: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description='Upload public comments',
        description='Upload the report public comments into the bug trackers',
        default=False,
    )
    upload_details_updates: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description='Upload details updates',
        description='Upload the report details updates into the bug trackers',
        default=False,
    )
    upload_rewards: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description='Upload rewards',
        description='Upload the report rewards into the bug trackers',
        default=False,
    )
    upload_status_updates: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description='Upload status updates',
        description='Upload the report status updates into the bug trackers',
        default=False,
    )

    def __init__(
        self,
        upload_private_comments: Optional[bool] = None,
        upload_public_comments: Optional[bool] = None,
        upload_details_updates: Optional[bool] = None,
        upload_rewards: Optional[bool] = None,
        upload_status_updates: Optional[bool] = None,
        **kwargs: Any,
    ):
        """
        Initialize self.

        Args:
            upload_private_comments: a flag indicating whether to upload private comments to the bugtrackers
            upload_public_comments: a flag indicating whether to upload public comments to the bugtrackers
            upload_details_updates: a flag indicating whether to upload details updates to the bugtrackers
            upload_rewards: a flag indicating whether to upload rewards to the bugtrackers
            upload_status_updates: a flag indicating whether to upload status updates to the bugtrackers
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.upload_private_comments = upload_private_comments
        self.upload_public_comments = upload_public_comments
        self.upload_details_updates = upload_details_updates
        self.upload_rewards = upload_rewards
        self.upload_status_updates = upload_status_updates


SynchronizeOptionsAttributeType = Union[
    Dict[str, bool],
    SynchronizeOptions,
    Attribute[SynchronizeOptions],
]
SynchronizeOptionsInitType = Union[
    Optional[SynchronizeOptions],
    Optional[Dict[str, bool]],
]


class Program(AttributesContainer):
    """A program and its associated bugtrackers."""

    slug: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='Program slug',
        description='Program slug',
        required=True,
        validator=not_blank_validator,
    )
    synchronize_options: SynchronizeOptionsAttributeType = Attribute.create(
        value_type=SynchronizeOptions,
        short_description='Synchronization options',
        required=True,
    )
    bugtrackers_name: BugtrackersNameAttributeType = Attribute.create(
        value_type=Bugtrackers,
        short_description='Bug trackers',
        description='Bug trackers names',
        required=True,
        validator=length_one_validator,
    )

    def __init__(
        self,
        slug: Optional[str] = None,
        synchronize_options: SynchronizeOptionsInitType = None,
        bugtrackers_name: BugtrackersNameInitType = None,
        **kwargs: Any,
    ):
        """
        Initialize the program.

        Args:
            slug: a program slug
            synchronize_options: synchronization options
            bugtrackers_name: a list of bugtrackers
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.slug = slug
        if not synchronize_options:
            synchronize_options = SynchronizeOptions()
        self.synchronize_options = synchronize_options
        if bugtrackers_name:
            if isinstance(bugtrackers_name, List):
                self.bugtrackers_name = Bugtrackers(bugtrackers_name)
            else:
                self.bugtrackers_name = cast(Bugtrackers, bugtrackers_name)


class Programs(AttributesContainerList[Program]):
    """A list of programs."""

    def __init__(
        self,
        items: Optional[List[Any]] = None,
    ) -> None:
        """
        Initialize the list of programs.

        Args:
            items: a list of programs
        """
        super().__init__(
            values_type=Program,
            items=items,
        )


AppHeadersAttributeType = Union[
    Optional[Dict[str, str]],
    Optional[Headers],
    Attribute[Headers],
]
AppHeadersInitType = Union[
    Optional[Headers],
    Optional[Dict[str, str]],
]
OAuthArgsAttributeType = Union[
    Optional[Dict[str, str]],
    Optional[OAuthSettings],
    Attribute[OAuthSettings],
]
OAuthArgsInitType = Union[
    Optional[OAuthSettings],
    Optional[Dict[str, str]],
]
ProgramsAttributeType = Union[
    List[Union[Program, Dict[str, Any]]],
    Optional[Programs],
    Attribute[Programs],
]
ProgramsInitType = Union[
    Optional[Programs],
    Optional[List[Union[Program, Dict[str, Any]]]],
]


class YesWeHackConfiguration(AttributesContainer):
    """A configuration for YesWeHack."""

    api_url: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='API URL',
        description='Base URL of the YWH API',
        default='https://apps.yeswehack.com',
        validator=url_validator,
    )
    apps_headers: AppHeadersAttributeType = Attribute.create(
        short_description='Apps headers',
        value_type=Headers,
        required=True,
        validator=dict_has_non_blank_key_validator(
            'X-YesWeHack-Apps',
        ),
    )
    login: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='Login',
        description='User login',
        required=True,
        validator=not_blank_validator,
    )
    password: StrAttributeType = Attribute.create(
        value_type=str,
        short_description='Password',
        description='User password',
        required=True,
        secret=True,
        validator=not_blank_validator,
    )
    oauth_args: OAuthArgsAttributeType = Attribute.create(
        value_type=OAuthSettings,
        short_description='OAuth settings',
        description='OAuth settings',
    )
    verify: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description='Verify TLS',
        description="Verify server's TLS certificate",
        default=True,
    )
    totp: BoolAttributeType = Attribute.create(
        value_type=bool,
        short_description='Use TOTP',
        description=''.join((
            'Use TOTP\n',
            "Apps API doesn't require TOTP authentication, even if corresponding user has TOTP enabled.\n",
            'However, on a secured program, information is limited for user with TOTP disabled, even in apps.\n',
            'As a consequence, to allow proper bug tracking integration on a secured program,',
            'program consumer must have TOTP enabled and, in BTI configuration TOTP must be set to false',
        )),
        default=True,
        deprecated=True,
    )
    programs: ProgramsAttributeType = Attribute.create(
        value_type=Programs,
        short_description='Programs',
        description='Programs',
        required=True,
        validator=not_empty_validator,
    )

    def __init__(
        self,
        api_url: Optional[str] = None,
        apps_headers: AppHeadersInitType = None,
        login: Optional[str] = None,
        password: Optional[str] = None,
        oauth_args: OAuthArgsInitType = None,
        verify: Optional[bool] = None,
        totp: Optional[bool] = None,
        programs: ProgramsInitType = None,
        **kwargs: Any,
    ):
        """
        Initialize the configuration.

        Args:
            api_url: a YesWeHack API URL
            apps_headers: a list of HTTP headers to be added to API calls
            login: a YesWeHack user login
            password: a YesWeHack user password
            oauth_args: settings for API OAuth authentication
            verify: a flag indicating whether to check SSL/TLS connection
            totp: a flag indicating whether to use TOTP
            programs: a list of Programs
            kwargs: keyword arguments
        """
        super().__init__(**kwargs)
        self.api_url = api_url
        self.apps_headers = apps_headers
        self.login = login
        self.password = password
        self.oauth_args = oauth_args
        self.verify = verify
        self.totp = totp
        self.programs = programs


class YesWeHackConfigurations(AttributesContainerDict[YesWeHackConfiguration]):
    """A list of YesWeHack configurations."""

    def __init__(
        self,
        **kwargs: Any,
    ):
        """
        Initialize the list of configurations.

        Args:
            kwargs: a list of configurations
        """
        super().__init__(
            values_type=YesWeHackConfiguration,
            items=kwargs,
        )



"""Models and functions used for formatting reports data."""

from abc import ABC, abstractmethod
from string import Template
from types import MappingProxyType
from typing import Optional

from singledispatchmethod import singledispatchmethod
from typing_extensions import Protocol

from ywh2bt.core.api.models.report import CommentLog, DetailsUpdateLog, Log, Report, RewardLog, StatusUpdateLog

PROPERTY_LABELS = MappingProxyType({
    'bug_type': 'Bug type',
    'scope': 'Scope',
    'end_point': 'Endpoint',
    'vulnerable_part': 'Vulnerable part',
    'part_name': 'Part name',
    'payload_sample': 'Payload',
    'application_finger_print': 'Application fingerprint',
})

STATUS_TRANSLATIONS = MappingProxyType({
    'new': 'New',
    'under_review': 'Under Review',
    'reopen_under_review': 'Reopen Under Review',
    'more_info': 'Need More Info',
    'reopen_more_info': 'Reopen Need More Info',
    'accepted': 'Accepted',
    'ask_verif': 'Ask for fix verification',
    'wont_fix': "Won't fix",
    'informative': 'Informative',
    'resolved': 'Resolved',
    'rtfs': 'RTFS',
    'spam': 'Spam',
    'out_of_scope': 'Out Of Scope',
    'not_applicable': 'Not Applicable',
    'invalid': 'Invalid',
    'duplicate': 'Duplicate',
    'auto_close': 'Auto Close',
})


class _ValueTransformer(Protocol):
    def __call__(
        self,
        value: str,
    ) -> str:
        ...  # noqa: WPS428


def _identity_transformer(
    value: str,
) -> str:
    return value


class ReportMessageFormatter(ABC):
    """A base report formatter."""

    _report_title_template: Template
    _report_description_template: Template
    _comment_body_template: Template
    _comment_log_template: Template
    _status_update_log_template: Template
    _details_update_log_template: Template
    _details_update_log_line_template: Template
    _reward_log_template: Template
    _value_transformer: _ValueTransformer

    def __init__(
        self,
        report_title_template: Template,
        report_description_template: Template,
        comment_body_template: Template,
        comment_log_template: Template,
        status_update_log_template: Template,
        details_update_log_template: Template,
        details_update_log_line_template: Template,
        reward_log_template: Template,
        value_transformer: Optional[_ValueTransformer] = None,
    ):
        """
        Initialize self.

        Args:
            report_title_template: a template for a report title
            report_description_template: a template for a report description
            comment_body_template: a template for an issue comment
            comment_log_template: a template for a template for a CommentLog
            status_update_log_template: a template for a StatusUpdateLog
            details_update_log_template: a template for a DetailsUpdateLog
            details_update_log_line_template: a template for entries of a DetailsUpdateLog
            reward_log_template: a template for a RewardLog
            value_transformer: a transformer for values
        """
        self._report_title_template = report_title_template
        self._report_description_template = report_description_template
        self._comment_body_template = comment_body_template
        self._comment_log_template = comment_log_template
        self._status_update_log_template = status_update_log_template
        self._details_update_log_template = details_update_log_template
        self._details_update_log_line_template = details_update_log_line_template
        self._reward_log_template = reward_log_template
        self._value_transformer = value_transformer or _identity_transformer

    def _transform_value(
        self,
        value: str,
    ) -> str:
        return self._value_transformer(
            value=value,
        )

    def format_report_title(
        self,
        report: Report,
    ) -> str:
        """
        Get a formatted title for a report.

        Args:
            report: a report

        Returns:
            a formatted title
        """
        return self._report_title_template.substitute(
            local_id=report.local_id,
            title=self._transform_value(
                value=report.title or '',
            ),
        )

    def format_report_description(
        self,
        report: Report,
    ) -> str:
        """
        Get a formatted description for a report.

        Args:
            report: a report

        Returns:
            a formatted description
        """
        return self._report_description_template.substitute(
            local_id=report.local_id,
            title=self._transform_value(
                value=report.title,
            ),
            priority_name=report.priority.name if report.priority else '',
            bug_type_label=self._get_property_label('bug_type'),
            bug_type_name=report.bug_type.name,
            bug_type_link=report.bug_type.link,
            bug_type_remediation_link=report.bug_type.remediation_link or '/',
            scope_label=self._get_property_label('scope'),
            scope=self._transform_value(
                value=report.scope,
            ),
            cvss_criticity=report.cvss.criticity,
            cvss_score=report.cvss.score,
            cvss_vector=report.cvss.vector,
            end_point_label=self._get_property_label('end_point'),
            end_point=self._transform_value(
                value=report.end_point,
            ),
            vulnerable_part_label=self._get_property_label('vulnerable_part'),
            vulnerable_part=self._transform_value(
                value=report.vulnerable_part,
            ),
            part_name_label=self._get_property_label('part_name'),
            part_name=self._transform_value(
                value=report.part_name,
            ),
            payload_sample_label=self._get_property_label('payload_sample'),
            payload_sample=self._transform_value(
                value=report.payload_sample or '',
            ),
            technical_information=self._transform_value(
                value=report.technical_information or '',
            ),
            description=self.transform_report_description_html(
                description_html=report.description_html,
            ),
        )

    def _get_property_label(
        self,
        property_name: str,
    ) -> str:
        return PROPERTY_LABELS.get(property_name, property_name)

    @abstractmethod
    def transform_report_description_html(
        self,
        description_html: str,
    ) -> str:
        """
        Transform the report description.

        Args:
            description_html: a report description in HTML
        """

    def format_log(
        self,
        log: Log,
    ) -> str:
        """
        Get a formatted comment for a log.

        Args:
            log: a log

        Returns:
            a formatted log
        """
        return self._comment_body_template.substitute(
            author=log.author.username,
            date=log.created_at,
            body=self._transform_log(
                log,
            ),
        )

    @abstractmethod
    def transform_html(
        self,
        html: str,
    ) -> str:
        """
        Transform html.

        Args:
            html: some HTML
        """

    @singledispatchmethod
    def _transform_log(
        self,
        log: Log,
    ) -> str:
        return self.transform_html(
            html=log.message_html,
        )

    @_transform_log.register
    def _transform_comment_log(
        self,
        log: CommentLog,
    ) -> str:
        return self._comment_log_template.substitute(
            comment=self.transform_html(
                html=log.message_html,
            ),
        )

    @_transform_log.register
    def _transform_status_update_log(
        self,
        log: StatusUpdateLog,
    ) -> str:
        return self._status_update_log_template.substitute(
            old_status=self._translate_status(
                status=(log.old_status or {}).get('workflow_state') or '',
            ),
            new_status=self._translate_status(
                status=(log.new_status or {}).get('workflow_state') or '',
            ),
            comment=self.transform_html(
                html=log.message_html,
            ),
        )

    def _translate_status(
        self,
        status: str,
    ) -> str:
        return STATUS_TRANSLATIONS.get(status, '')

    @_transform_log.register
    def _transform_details_update_log(
        self,
        log: DetailsUpdateLog,
    ) -> str:
        details_lines = []
        for updated_property, new_value in (log.new_details or {}).items():
            old_value = (log.old_details or {}).get(updated_property, '')
            details_lines.append(
                self._details_update_log_line_template.substitute(
                    updated_property=self._get_property_label(
                        property_name=updated_property,
                    ),
                    old_value=self._transform_value(
                        value=old_value or '',
                    ),
                    new_value=self._transform_value(
                        value=new_value or '',
                    ),
                ),
            )
        return self._details_update_log_template.substitute(
            details_lines=''.join(details_lines),
        )

    @_transform_log.register
    def _format_reward_log(
        self,
        log: RewardLog,
    ) -> str:
        return self._reward_log_template.substitute(
            comment=self.transform_html(
                html=log.message_html,
            ),
        )



"""Models and functions used for data synchronisation between YesWeHack and GitLab trackers."""
from __future__ import annotations

from typing import Iterable, List, Optional, Tuple, cast

import requests
from gitlab import Gitlab, GitlabError  # type: ignore
from gitlab.v4.objects import Project, ProjectIssue, ProjectIssueNote  # type: ignore

from ywh2bt.core.api.formatter.markdown import ReportMessageMarkdownFormatter
from ywh2bt.core.api.models.report import Attachment, Log, Report
from ywh2bt.core.api.tracker import (
    TrackerClient,
    TrackerClientError,
    TrackerComment,
    TrackerComments,
    TrackerIssue,
)
from ywh2bt.core.configuration.trackers.gitlab import GitLabConfiguration


class GitLabTrackerClientError(TrackerClientError):
    """A GitLab tracker client error."""


class GitLabTrackerClient(TrackerClient[GitLabConfiguration]):
    """A GitLab tracker client."""

    _session: requests.Session
    _gitlab: Gitlab
    _message_formatter: ReportMessageMarkdownFormatter

    def __init__(
        self,
        configuration: GitLabConfiguration,
    ):
        """
        Initialize self.

        Args:
            configuration: a GitLab configuration
        """
        super().__init__(
            configuration=configuration,
        )
        self._session = requests.Session()
        self._session.verify = configuration.verify
        self._gitlab = Gitlab(
            url=configuration.url,
            private_token=configuration.token,
            session=self._session,
        )
        self._message_formatter = ReportMessageMarkdownFormatter()

    @property
    def tracker_type(self) -> str:
        """
        Get the type of the tracker client.

        Returns:
            the type of the  tracker client
        """
        return 'GitLab'

    def _build_tracker_issue(
        self,
        issue_id: str,
        issue_url: str,
        is_closed: bool,
    ) -> TrackerIssue:
        return TrackerIssue(
            tracker_url=cast(str, self.configuration.url),
            project=cast(str, self.configuration.project),
            issue_id=issue_id,
            issue_url=issue_url,
            is_closed=is_closed,
        )

    def get_tracker_issue(
        self,
        issue_id: str,
    ) -> Optional[TrackerIssue]:
        """
        Get a tracker issue.

        Args:
            issue_id: an issue id

        Returns:
            The issue if it exists, else None
        """
        try:
            gitlab_issue = self._get_gitlab_issue(
                issue_id=issue_id,
            )
        except GitLabTrackerClientError:
            return None
        if gitlab_issue is None:
            return None
        return self._build_tracker_issue(
            issue_id=issue_id,
            issue_url=gitlab_issue.web_url,
            is_closed=gitlab_issue.state == 'closed',
        )

    def send_report(
        self,
        report: Report,
    ) -> TrackerIssue:
        """
        Send a report to the tracker.

        Args:
            report: a report

        Returns:
            information about the sent report
        """
        self._ensure_auth()
        gitlab_project = self._get_gitlab_project()
        description = self._message_formatter.format_report_description(
            report=report,
        )
        uploads = self._upload_attachments(
            gitlab_project=gitlab_project,
            attachments=report.attachments,
        )
        for attachment_url, uploaded_url in uploads:
            description = description.replace(
                attachment_url,
                uploaded_url,
            )
        gitlab_issue = self._create_issue(
            gitlab_project=gitlab_project,
            title=self._message_formatter.format_report_title(
                report=report,
            ),
            description=description,
        )
        return self._build_tracker_issue(
            issue_id=gitlab_issue.id,
            issue_url=gitlab_issue.web_url,
            is_closed=False,
        )

    def send_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: Iterable[Log],
    ) -> TrackerComments:
        """
        Send logs to the tracker.

        Args:
            tracker_issue: information about the tracker issue
            logs: a list of comments

        Raises:
            GitLabTrackerClientError: if an error occurs

        Returns:
            information about the sent comments
        """
        self._ensure_auth()
        tracker_comments = TrackerComments(
            tracker_issue=tracker_issue,
            added_comments=[],
        )
        gitlab_project = self._get_gitlab_project()
        gitlab_issue = self._get_gitlab_issue(
            issue_id=tracker_issue.issue_id,
        )
        if not gitlab_issue:
            raise GitLabTrackerClientError(
                f'GitLab issue {tracker_issue.issue_id} not found in project {self.configuration.project}',
            )
        for log in logs:
            gitlab_comment = self._add_comment(
                gitlab_project=gitlab_project,
                gitlab_issue=gitlab_issue,
                log=log,
            )
            tracker_comments.added_comments.append(
                TrackerComment(
                    comment_id=gitlab_comment.id,
                ),
            )
        return tracker_comments

    def test(
        self,
    ) -> None:
        """
        Test the client.

        Raises:
            GitLabTrackerClientError: if the test failed
        """
        try:
            self._gitlab.auth()
        except GitlabError as e:
            raise GitLabTrackerClientError('Unable to log in with GitLab API client') from e

    def _get_gitlab_project(
        self,
    ) -> Project:
        try:
            return self._gitlab.projects.get(
                self.configuration.project,
            )
        except GitlabError as e:
            raise GitLabTrackerClientError(
                f'Unable to get GitLab project {self.configuration.project}',
            ) from e

    def _get_gitlab_issue(
        self,
        issue_id: str,
    ) -> Optional[ProjectIssue]:
        gitlab_project = self._get_gitlab_project()
        try:
            gitlab_issues = gitlab_project.issues.list(all=False, as_list=False)
        except GitlabError as e:
            raise GitLabTrackerClientError(
                f'Unable to get GitLab issues for project {self.configuration.project}',
            ) from e
        issue_id_int = int(issue_id)
        for gitlab_issue in gitlab_issues:
            if gitlab_issue.id == issue_id_int:
                return gitlab_issue
        return None

    def _upload_attachments(
        self,
        gitlab_project: Project,
        attachments: List[Attachment],
    ) -> List[Tuple[str, str]]:
        try:
            return [
                (
                    attachment.url,
                    gitlab_project.upload(attachment.original_name, attachment.data)['url'],
                )
                for attachment in attachments
            ]
        except GitlabError as e:
            raise GitLabTrackerClientError(
                f'Unable to upload attachments for project {self.configuration.project} to GitLab',
            ) from e

    def _create_issue(
        self,
        gitlab_project: Project,
        title: str,
        description: str,
    ) -> ProjectIssue:
        issue_data = {
            'title': title,
            'description': description,
        }
        try:
            return gitlab_project.issues.create(issue_data)
        except GitlabError as e:
            raise GitLabTrackerClientError(
                f'Unable to create issue for project {self.configuration.project} to GitLab',
            ) from e

    def _add_comment(
        self,
        gitlab_project: Project,
        gitlab_issue: ProjectIssue,
        log: Log,
    ) -> ProjectIssueNote:
        comment_body = self._message_formatter.format_log(
            log=log,
        )
        uploads = self._upload_attachments(
            gitlab_project=gitlab_project,
            attachments=log.attachments,
        )
        for attachment_url, uploaded_url in uploads:
            comment_body = comment_body.replace(
                attachment_url,
                uploaded_url,
            )
        try:
            return gitlab_issue.notes.create({
                'body': comment_body,
            })
        except GitlabError as e:
            raise GitLabTrackerClientError(
                f'Unable to add GitLab comment for issue {gitlab_issue} in project {self.configuration.project}',
            ) from e

    def _ensure_auth(
        self,
    ) -> None:
        try:
            self._gitlab.auth()
        except GitlabError as e:
            raise GitLabTrackerClientError('Unable to authenticate to GitLab') from e



"""Models and functions used for tracker clients mixins."""
from typing import Any, Dict

from ywh2bt.core.api.tracker import TrackerClient, TrackerClientError
from ywh2bt.core.configuration.tracker import TrackerConfiguration
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.mapping import TRACKER_CLIENT_CLASSES


class TrackerClientsMixin:
    """Mixin for TrackerClients."""

    _tracker_clients: Dict[TrackerConfiguration, TrackerClient[Any]]

    def __init__(
        self,
    ) -> None:
        """Initialize self."""
        self._tracker_clients = {}

    def get_tracker_client(
        self,
        configuration: TrackerConfiguration,
    ) -> TrackerClient[Any]:
        """
        Get the api client for the given configuration.

        Args:
            configuration: a configuration

        Returns:
            The client
        """
        if configuration not in self._tracker_clients:
            self._tracker_clients[configuration] = self._build_tracker_client(
                configuration=configuration,
            )
        return self._tracker_clients[configuration]

    def _build_tracker_client(
        self,
        configuration: TrackerConfiguration,
    ) -> TrackerClient[Any]:
        configuration_class = configuration.__class__
        if configuration_class not in TRACKER_CLIENT_CLASSES:
            raise CoreException(f'{configuration_class} does not have an associated tracker client')
        tracker_client_class = TRACKER_CLIENT_CLASSES[configuration_class]
        try:
            client = tracker_client_class(
                configuration=configuration,
            )
        except TrackerClientError as e:
            raise CoreException(f'Unable to create {configuration_class} API client') from e
        return client



"""Models and functions used for YesWeHack api clients mixins."""
from typing import Dict

from ywh2bt.core.api.yeswehack import YesWeHackApiClient, YesWeHackApiClientError
from ywh2bt.core.configuration.yeswehack import YesWeHackConfiguration
from ywh2bt.core.exceptions import CoreException


class YesWeHackApiClientsMixin:
    """Mixin for YesWeHackApiClients."""

    _yeswehack_api_clients: Dict[YesWeHackConfiguration, YesWeHackApiClient]

    def __init__(
        self,
    ) -> None:
        """Initialize self."""
        self._yeswehack_api_clients = {}

    def get_yeswehack_api_client(
        self,
        configuration: YesWeHackConfiguration,
    ) -> YesWeHackApiClient:
        """
        Get the api client for the given configuration.

        Args:
            configuration: a configuration

        Raises:
            CoreException: if the client could not be created

        Returns:
            The client
        """
        if configuration not in self._yeswehack_api_clients:
            try:
                client = YesWeHackApiClient(
                    configuration=configuration,
                )
            except YesWeHackApiClientError as e:
                raise CoreException('Unable to create YesWeHack API client') from e
            self._yeswehack_api_clients[configuration] = client
        return self._yeswehack_api_clients[configuration]



"""Models and functions used for data synchronisation between YesWeHack and trackers."""
from __future__ import annotations

from string import Template
from typing import Any, List, Optional, Tuple, cast

from ywh2bt.core.api.models.report import (
    CommentLog,
    DetailsUpdateLog,
    Log,
    Report,
    RewardLog,
    StatusUpdateLog,
    TrackerUpdateLog,
    TrackingStatusLog,
)
from ywh2bt.core.api.tracker import TrackerClient, TrackerClientError, TrackerComments, TrackerIssue, TrackerIssueState
from ywh2bt.core.api.yeswehack import YesWeHackApiClient, YesWeHackApiClientError
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.configuration.tracker import Trackers
from ywh2bt.core.configuration.yeswehack import (
    Program,
    Programs,
    SynchronizeOptions,
    YesWeHackConfiguration,
    YesWeHackConfigurations,
)
from ywh2bt.core.mixins.tracker_clients import TrackerClientsMixin
from ywh2bt.core.mixins.yeswehack_api_clients import YesWeHackApiClientsMixin
from ywh2bt.core.state.decrypt import StateDecryptor
from ywh2bt.core.state.encrypt import StateEncryptor
from ywh2bt.core.synchronizer.error import SynchronizerError
from ywh2bt.core.synchronizer.listener import (
    NoOpSynchronizerListener,
    SynchronizerEndEvent,
    SynchronizerEndFetchReportsEvent,
    SynchronizerEndSendReportEvent,
    SynchronizerEvent,
    SynchronizerListener,
    SynchronizerStartEvent,
    SynchronizerStartFetchReportsEvent,
    SynchronizerStartSendReportEvent,
)


class Synchronizer(YesWeHackApiClientsMixin, TrackerClientsMixin):  # noqa: WPS214
    """A class used for data synchronisation between YesWeHack and trackers."""

    _configuration: RootConfiguration
    _listener: SynchronizerListener
    _message_formatter: _MessageFormatter

    def __init__(
        self,
        configuration: RootConfiguration,
        listener: Optional[SynchronizerListener] = None,
    ):
        """
        Initialize self.

        Args:
            configuration: a configuration
            listener: an observer that will receive synchronization events
        """
        YesWeHackApiClientsMixin.__init__(self)  # noqa: WPS609  # type: ignore
        TrackerClientsMixin.__init__(self)  # noqa: WPS609  # type: ignore
        self._configuration = configuration
        self._listener = listener or NoOpSynchronizerListener()
        self._message_formatter = _MessageFormatter()

    def synchronize(
        self,
    ) -> None:
        """Synchronize YesWeHack and the trackers."""
        self._send_event(
            event=SynchronizerStartEvent(
                configuration=self._configuration,
            ),
        )
        yeswehack_configurations = cast(YesWeHackConfigurations, self._configuration.yeswehack)
        for yeswehack_name, yeswehack_configuration in yeswehack_configurations.items():
            self._send_programs_from_yeswehack_configuration(
                configuration=self._configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
            )
        self._send_event(
            event=SynchronizerEndEvent(
                configuration=self._configuration,
            ),
        )

    def _send_event(
        self,
        event: SynchronizerEvent,
    ) -> None:
        self._listener.on_event(
            event=event,
        )

    def _send_programs_from_yeswehack_configuration(
        self,
        configuration: RootConfiguration,
        yeswehack_name: str,
        yeswehack_configuration: YesWeHackConfiguration,
    ) -> None:
        yeswehack_client = self.get_yeswehack_api_client(
            configuration=yeswehack_configuration,
        )
        programs = cast(Programs, yeswehack_configuration.programs)
        for program in programs:
            if program.slug is None:
                continue
            self._send_event(
                event=SynchronizerStartFetchReportsEvent(
                    configuration=configuration,
                    yeswehack_name=yeswehack_name,
                    yeswehack_configuration=yeswehack_configuration,
                    program_slug=program.slug,
                ),
            )
            reports = self._get_afi_reports(
                yeswehack_client=yeswehack_client,
                program=program,
            )
            self._send_event(
                event=SynchronizerEndFetchReportsEvent(
                    configuration=configuration,
                    yeswehack_name=yeswehack_name,
                    yeswehack_configuration=yeswehack_configuration,
                    program_slug=program.slug,
                    reports=reports,
                ),
            )
            self._send_reports_to_trackers(
                configuration=configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
                yeswehack_client=yeswehack_client,
                program=program,
                reports=reports,
            )

    def _get_afi_reports(
        self,
        yeswehack_client: YesWeHackApiClient,
        program: Program,
    ) -> List[Report]:
        program_slug = cast(str, program.slug)
        filters = {
            'filter[trackingStatus][0]': 'AFI',
        }
        synchronize_options = cast(SynchronizeOptions, program.synchronize_options)
        include_tracked = any(
            (
                synchronize_options.upload_private_comments,
                synchronize_options.upload_public_comments,
                synchronize_options.upload_details_updates,
                synchronize_options.upload_rewards,
                synchronize_options.upload_status_updates,
            ),
        )
        if include_tracked:
            filters['filter[trackingStatus][1]'] = 'T'
        try:
            return yeswehack_client.get_program_reports(
                slug=program_slug,
                filters=filters,
            )
        except YesWeHackApiClientError as e:
            raise SynchronizerError(
                f'Unable to get AFI/T reports for program {program_slug}',
            ) from e

    def _send_reports_to_trackers(
        self,
        configuration: RootConfiguration,
        yeswehack_name: str,
        yeswehack_configuration: YesWeHackConfiguration,
        yeswehack_client: YesWeHackApiClient,
        program: Program,
        reports: List[Report],
    ) -> None:
        for report in reports:
            self._send_report_to_trackers(
                configuration=configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
                yeswehack_client=yeswehack_client,
                program=program,
                report=report,
            )

    def _send_report_to_trackers(
        self,
        configuration: RootConfiguration,
        yeswehack_name: str,
        yeswehack_configuration: YesWeHackConfiguration,
        yeswehack_client: YesWeHackApiClient,
        program: Program,
        report: Report,
    ) -> None:
        if not program.bugtrackers_name:
            return
        for bugtracker_name in program.bugtrackers_name:
            self._send_report_to_tracker(
                configuration=configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
                yeswehack_client=yeswehack_client,
                bugtracker_name=bugtracker_name,
                synchronize_options=cast(SynchronizeOptions, program.synchronize_options),
                report=report,
            )

    def _send_report_to_tracker(  # noqa: WPS210
        self,
        configuration: RootConfiguration,
        yeswehack_name: str,
        yeswehack_configuration: YesWeHackConfiguration,
        yeswehack_client: YesWeHackApiClient,
        bugtracker_name: str,
        synchronize_options: SynchronizeOptions,
        report: Report,
    ) -> None:
        bugtracker_configuration = cast(Trackers, self._configuration.bugtrackers)[bugtracker_name]
        tracker_client = self.get_tracker_client(
            configuration=bugtracker_configuration,
        )
        self._send_event(
            event=SynchronizerStartSendReportEvent(
                configuration=configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
                program_slug=report.program.slug,
                tracker_name=bugtracker_name,
                report=report,
            ),
        )
        tracker_issue = self._get_tracker_issue_from_logs(
            bugtracker_name=bugtracker_name,
            tracker_client=tracker_client,
            report=report,
        )
        is_existing_issue = tracker_issue is not None
        if not is_existing_issue:
            tracker_issue, _ = self._synchronize_tracker_issue_and_get_logs(
                bugtracker_name=bugtracker_name,
                tracker_client=tracker_client,
                report=report,
            )
        if tracker_issue is None:
            raise SynchronizerError(
                f'Unable to create new issue or get existing issue for #{report.report_id} in {bugtracker_name}',
            )
        if report.tracking_status != 'T' or not is_existing_issue:
            self._update_tracking_status(
                yeswehack_client=yeswehack_client,
                bugtracker_name=bugtracker_name,
                tracker_client=tracker_client,
                report=report,
                tracker_issue=tracker_issue,
            )
        logs = report.logs
        if is_existing_issue:
            last_tracker_update_log = self._get_last_tracker_update_log(
                report=report,
                bugtracker_name=bugtracker_name,
            )
            if last_tracker_update_log:
                logs = report.get_logs_after(
                    log=last_tracker_update_log,
                )
        tracker_comments = self._send_synchronizable_logs(
            bugtracker_name=bugtracker_name,
            tracker_client=tracker_client,
            tracker_issue=tracker_issue,
            synchronize_options=synchronize_options,
            logs=logs,
            report=report,
        )
        if tracker_comments.added_comments:
            yeswehack_client.post_report_tracker_update(
                report=report,
                tracker_name=bugtracker_name,
                issue_id=tracker_issue.issue_id,
                issue_url=tracker_issue.issue_url,
                token=StateEncryptor.encrypt(
                    key=report.report_id,
                    state=TrackerIssueState(
                        closed=False,
                        bugtracker_name=bugtracker_name,
                    ),
                ),
                comment=self._message_formatter.format_synchronization_done_message(
                    tracker_type=tracker_client.tracker_type,
                    report=report,
                    tracker_comments=tracker_comments,
                ),
            )
        self._send_event(
            event=SynchronizerEndSendReportEvent(
                configuration=configuration,
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
                program_slug=report.program.slug,
                tracker_name=bugtracker_name,
                report=report,
                is_existing_issue=is_existing_issue,
                tracking_status_updated=report.tracking_status != 'T' or not is_existing_issue,
                result=tracker_comments,
            ),
        )

    def _get_last_tracker_update_log(
        self,
        bugtracker_name: str,
        report: Report,
    ) -> Optional[Log]:
        for log in reversed(report.logs):
            if isinstance(log, TrackerUpdateLog):
                state = StateDecryptor.decrypt(
                    encrypted_state=log.tracker_token or '',
                    key=report.report_id,
                    state_type=TrackerIssueState,
                )
                if state and state.bugtracker_name == bugtracker_name:
                    return log
        return None

    def _send_synchronizable_logs(
        self,
        bugtracker_name: str,
        tracker_client: TrackerClient[Any],
        tracker_issue: TrackerIssue,
        synchronize_options: SynchronizeOptions,
        logs: List[Log],
        report: Report,
    ) -> TrackerComments:
        synchronizable_logs = []
        for log in logs:
            synchronize = self._is_synchronizable_log(
                synchronize_options=synchronize_options,
                log=log,
            )
            if synchronize:
                synchronizable_logs.append(log)
        if not synchronizable_logs:
            return TrackerComments(
                tracker_issue=tracker_issue,
                added_comments=[],
            )
        return self._send_logs(
            bugtracker_name=bugtracker_name,
            tracker_client=tracker_client,
            tracker_issue=tracker_issue,
            logs=synchronizable_logs,
            report=report,
        )

    def _is_synchronizable_log(
        self,
        synchronize_options: SynchronizeOptions,
        log: Log,
    ) -> bool:
        return any((
            isinstance(log, CommentLog) and synchronize_options.upload_public_comments and not log.private,
            isinstance(log, CommentLog) and synchronize_options.upload_private_comments and log.private,
            isinstance(log, DetailsUpdateLog) and synchronize_options.upload_details_updates,
            isinstance(log, RewardLog) and synchronize_options.upload_rewards,
            isinstance(log, StatusUpdateLog) and synchronize_options.upload_status_updates,
        ))

    def _send_logs(
        self,
        bugtracker_name: str,
        tracker_client: TrackerClient[Any],
        tracker_issue: TrackerIssue,
        logs: List[Log],
        report: Report,
    ) -> TrackerComments:
        try:
            return tracker_client.send_logs(
                tracker_issue=tracker_issue,
                logs=logs,
            )
        except TrackerClientError as send_error:
            raise SynchronizerError(
                f'Unable to send logs for #{report.report_id} to {bugtracker_name}',
            ) from send_error

    def _update_tracking_status(
        self,
        yeswehack_client: YesWeHackApiClient,
        bugtracker_name: str,
        tracker_client: TrackerClient[Any],
        report: Report,
        tracker_issue: TrackerIssue,
    ) -> None:
        message = self._message_formatter.format_tracking_status_update_message(
            tracker_type=tracker_client.tracker_type,
            tracker_issue=tracker_issue,
        )
        try:
            yeswehack_client.put_report_tracking_status(
                report=report,
                status='T',
                tracker_name=bugtracker_name,
                issue_id=tracker_issue.issue_id,
                issue_url=tracker_issue.issue_url,
                comment=message,
            )
        except YesWeHackApiClientError as tracking_status_error:
            raise SynchronizerError(
                f'Unable to update tracking status for report #{report.report_id}',
            ) from tracking_status_error

    def _get_tracker_issue_from_logs(
        self,
        bugtracker_name: str,
        tracker_client: TrackerClient[Any],
        report: Report,
    ) -> Optional[TrackerIssue]:
        log = report.get_last_tracking_status_update_log(
            tracker_name=bugtracker_name,
        )
        if log and isinstance(log, TrackingStatusLog) and all((log.tracker_id, log.tracker_url)):
            return tracker_client.get_tracker_issue(
                issue_id=cast(str, log.tracker_id),
            )
        return None

    def _synchronize_tracker_issue_and_get_logs(
        self,
        bugtracker_name: str,
        tracker_client: TrackerClient[Any],
        report: Report,
    ) -> Tuple[Optional[TrackerIssue], List[Log]]:
        try:
            tracker_issue = tracker_client.send_report(
                report=report,
            )
        except TrackerClientError as send_error:
            raise SynchronizerError(
                f'Unable to send report #{report.report_id} to {bugtracker_name}',
            ) from send_error
        return (
            tracker_issue,
            report.get_comments(),
        )


class _MessageFormatter:
    _tracking_status_update_template: Template = Template(
        'Synchronized with bugtracker : ${tracker_url} on project : ${project}.'
        + '\n'
        + 'Tracked to [${tracker_type} #${issue_id}](${issue_url}).',
    )
    _synchronization_done_template: Template = Template(
        'Synchronized with bugtracker : ${tracker_url} on project : ${project}.'
        + '\n'
        + 'Tracked to [${tracker_type} #${issue_id}](${issue_url}).'
        + '\n'
        + 'Report comments added to issue: ${comment_count}',
    )

    def format_tracking_status_update_message(
        self,
        tracker_type: str,
        tracker_issue: TrackerIssue,
    ) -> str:
        return self._tracking_status_update_template.substitute(
            tracker_type=tracker_type,
            tracker_url=tracker_issue.tracker_url,
            project=tracker_issue.project,
            issue_id=tracker_issue.issue_id,
            issue_url=tracker_issue.issue_url,
        )

    def format_synchronization_done_message(
        self,
        report: Report,
        tracker_type: str,
        tracker_comments: TrackerComments,
    ) -> str:
        tracker_issue = tracker_comments.tracker_issue
        return self._synchronization_done_template.substitute(
            tracker_type=tracker_type,
            tracker_url=tracker_issue.tracker_url,
            project=tracker_issue.project,
            issue_id=tracker_issue.issue_id,
            issue_url=tracker_issue.issue_url,
            comment_count=len(tracker_comments.added_comments),
        )

    def _string_state(
        self,
        is_existing_issue: bool,
        issue_state: TrackerIssueState,
        last_state: Optional[TrackerIssueState],
    ) -> str:
        if is_existing_issue:
            return self._string_state_for_update(
                issue_state=issue_state,
                last_state=last_state,
            )
        return 'created'

    def _string_state_for_update(
        self,
        issue_state: TrackerIssueState,
        last_state: Optional[TrackerIssueState],
    ) -> str:
        string_state = []
        if issue_state != last_state:
            if last_state:
                string_state.append('closed' if last_state.closed else 'opened')
            else:
                string_state.append('???')
            string_state.append(' -> ')
            string_state.append('closed' if issue_state.closed else 'opened')
        return ''.join(string_state)



"""Models and functions used for data synchronisation between YesWeHack and trackers."""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, List, Optional, TypeVar

from ywh2bt.core.api.client import TestableApiClient
from ywh2bt.core.api.models.report import Log, Report
from ywh2bt.core.configuration.tracker import TrackerConfiguration
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.state.state import State


class TrackerClientError(CoreException):
    """A base tracker client error."""


@dataclass
class TrackerIssueState(State):
    """A state of an issue."""

    closed: bool = False
    bugtracker_name: Optional[str] = None

    def __eq__(
        self,
        other: Optional[object],
    ) -> bool:
        """
        Indicate if 2 states are the same.

        Args:
            other: a state

        Returns:
            True if the states are the same
        """
        return isinstance(other, TrackerIssueState) and all(
            (
                self.closed == other.closed,
                self.bugtracker_name == other.bugtracker_name,
            ),
        )


@dataclass
class TrackerIssue:
    """Information about a tracker issue."""

    tracker_url: str
    project: str
    issue_id: str
    issue_url: str
    is_closed: bool


@dataclass
class TrackerComment:
    """Information about a tracker comment on an issue."""

    comment_id: str


@dataclass
class TrackerComments:
    """A result of sending comments to a tracker."""

    tracker_issue: TrackerIssue
    added_comments: List[TrackerComment]


T = TypeVar('T', bound=TrackerConfiguration)


class TrackerClient(TestableApiClient, ABC, Generic[T]):
    """An abstract base tracker client."""

    configuration: T

    def __init__(
        self,
        configuration: T,
    ):
        """
        Initialize self.

        Args:
            configuration: a tracker configuration
        """
        self.configuration = configuration

    @property
    @abstractmethod
    def tracker_type(self) -> str:
        """Get the type of the tracker client."""

    @abstractmethod
    def get_tracker_issue(
        self,
        issue_id: str,
    ) -> Optional[TrackerIssue]:
        """
        Get a tracker issue.

        Args:
            issue_id: an issue id
        """

    @abstractmethod
    def send_report(
        self,
        report: Report,
    ) -> TrackerIssue:
        """
        Send a report to the tracker.

        Args:
            report: a report
        """

    @abstractmethod
    def send_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: List[Log],
    ) -> TrackerComments:
        """
        Send logs to the tracker.

        Args:
            tracker_issue: information about the tracker issue
            logs: a list of comments
        """



"""Models and functions used for data synchronisation between YesWeHack and GitHub trackers."""
import re
from string import Template
from typing import Iterable, List, Optional, cast

from github import Github, GithubException
from github.Issue import Issue
from github.IssueComment import IssueComment
from github.Repository import Repository
from requests import RequestException

from ywh2bt.core.api.formatter.markdown import ReportMessageMarkdownFormatter
from ywh2bt.core.api.models.report import Attachment, Log, Report
from ywh2bt.core.api.tracker import (
    TrackerClient,
    TrackerClientError,
    TrackerComment,
    TrackerComments,
    TrackerIssue,
)
from ywh2bt.core.api.trackers.github.attachment import GitHubAttachmentUploader
from ywh2bt.core.configuration.trackers.github import GitHubConfiguration


class GitHubTrackerClientError(TrackerClientError):
    """A GitHub tracker client error."""


class GitHubTrackerClient(TrackerClient[GitHubConfiguration]):
    """A GitHub tracker client."""

    _attachment_name_regex_template = Template(r'(?:!?\[)([^\[\]]*)(?:\])(?:\(${url}\))')
    _attachment_substitute_regex_template = Template(r'!?\[[^\[\]]*\]\(${url}\)')
    _github: Github
    _attachment_uploader: GitHubAttachmentUploader
    _message_formatter: ReportMessageMarkdownFormatter

    def __init__(
        self,
        configuration: GitHubConfiguration,
    ):
        """
        Initialize self.

        Args:
            configuration: a GitHub configuration
        """
        super().__init__(
            configuration=configuration,
        )
        self._github = Github(
            base_url=cast(str, self.configuration.url),
            login_or_token=cast(str, self.configuration.token),
            verify=cast(bool, self.configuration.verify),
        )
        self._attachment_uploader = GitHubAttachmentUploader(
            configuration=self.configuration,
        )
        self._message_formatter = ReportMessageMarkdownFormatter()

    @property
    def tracker_type(self) -> str:
        """
        Get the type of the tracker client.

        Returns:
            the type of the  tracker client
        """
        return 'GitHub'

    def _build_tracker_issue(
        self,
        issue_id: str,
        issue_url: str,
        is_closed: bool,
    ) -> TrackerIssue:
        return TrackerIssue(
            tracker_url=cast(str, self.configuration.url),
            project=cast(str, self.configuration.project),
            issue_id=issue_id,
            issue_url=issue_url,
            is_closed=is_closed,
        )

    def get_tracker_issue(
        self,
        issue_id: str,
    ) -> Optional[TrackerIssue]:
        """
        Get a tracker issue.

        Args:
            issue_id: an issue id

        Returns:
            The issue if it exists, else None
        """
        try:
            github_issue = self._get_github_issue(
                issue_id=issue_id,
            )
        except GitHubTrackerClientError:
            return None
        if github_issue is None:
            return None
        return self._build_tracker_issue(
            issue_id=issue_id,
            issue_url=github_issue.html_url,
            is_closed=github_issue.closed_at is not None,
        )

    def send_report(
        self,
        report: Report,
    ) -> TrackerIssue:
        """
        Send a report to the tracker.

        Args:
            report: a report

        Returns:
            information about the sent report
        """
        self._ensure_auth()
        repository = self._get_repository()
        title = self._message_formatter.format_report_title(
            report=report,
        )
        body = self._message_formatter.format_report_description(
            report=report,
        )
        issue = self._create_issue(
            github_repository=repository,
            title=title,
            body=body,
        )
        self._upload_issue_attachments(
            issue=issue,
            attachments=report.attachments,
        )
        return self._build_tracker_issue(
            issue_id=str(issue.id),
            issue_url=issue.html_url,
            is_closed=False,
        )

    def send_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: Iterable[Log],
    ) -> TrackerComments:
        """
        Send logs to the tracker.

        Args:
            tracker_issue: information about the tracker issue
            logs: a list of comments

        Raises:
            GitHubTrackerClientError: if an error occurs

        Returns:
            information about the sent comments
        """
        tracker_comments = TrackerComments(
            tracker_issue=tracker_issue,
            added_comments=[],
        )
        github_issue = self._get_github_issue(
            issue_id=tracker_issue.issue_id,
        )
        if not github_issue:
            raise GitHubTrackerClientError(f'GitHub issue {tracker_issue.issue_id} not found')
        for log in logs:
            github_comment = self._add_comment(
                github_issue=github_issue,
                log=log,
            )
            self._upload_comment_attachments(
                issue=github_issue,
                comment=github_comment,
                attachments=log.attachments,
            )
            tracker_comments.added_comments.append(
                TrackerComment(
                    comment_id=str(github_comment.id),
                ),
            )
        return tracker_comments

    def test(
        self,
    ) -> None:
        """
        Test the client.

        Raises:
            GitHubTrackerClientError: if the test failed
        """
        try:
            login = self._github.get_user().login
        except GithubException as e:
            raise GitHubTrackerClientError('Unable to log in with GitHub API client') from e
        if not login:
            raise GitHubTrackerClientError('Unable to log in with GitHub API client')

    def _get_repository(
        self,
    ) -> Repository:
        try:
            return self._github.get_repo(
                cast(str, self.configuration.project),
            )
        except (GithubException, RequestException) as e:
            raise GitHubTrackerClientError(
                f'Unable to get GitHub repository {self.configuration.project}',
            ) from e

    def _create_issue(
        self,
        github_repository: Repository,
        title: str,
        body: str,
    ) -> Issue:
        try:
            return github_repository.create_issue(
                title=title,
                body=body,
            )
        except GithubException as e:
            raise GitHubTrackerClientError(
                f'Unable to create issue for project {self.configuration.project} to GitHub',
            ) from e

    def _add_comment(
        self,
        github_issue: Issue,
        log: Log,
    ) -> IssueComment:
        comment_body = self._message_formatter.format_log(
            log=log,
        )
        try:
            return github_issue.create_comment(
                body=comment_body,
            )
        except GithubException as e:
            raise GitHubTrackerClientError(
                f'Unable to add GitHub comment for issue {github_issue} in project {self.configuration.project}',
            ) from e

    def _get_github_issue(
        self,
        issue_id: str,
    ) -> Optional[Issue]:
        github_repository = self._get_repository()
        issue_id_int = int(issue_id)
        try:
            for issue in github_repository.get_issues(state='all'):
                if issue.id == issue_id_int:
                    return issue
        except GithubException as e:
            raise GitHubTrackerClientError(
                f'GitHub issue {issue_id} not found in project {self.configuration.project}',
            ) from e
        return None

    def _upload_issue_attachments(
        self,
        issue: Issue,
        attachments: List[Attachment],
    ) -> None:
        body = self._upload_attachments(
            issue=issue,
            body=issue.body,
            attachments=attachments,
        )
        issue.edit(
            body=body,
        )

    def _upload_comment_attachments(
        self,
        issue: Issue,
        comment: IssueComment,
        attachments: List[Attachment],
    ) -> None:
        body = self._upload_attachments(
            issue=issue,
            body=comment.body,
            attachments=attachments,
        )
        comment.edit(
            body=body,
        )

    def _upload_attachments(
        self,
        issue: Issue,
        body: str,
        attachments: List[Attachment],
    ) -> str:
        for attachment in attachments:
            attachment_name = self._extract_attachment_name(
                body=body,
                attachment=attachment,
            )
            if self.configuration.github_cdn_on:
                url = self._attachment_uploader.upload_attachment(
                    issue=issue,
                    attachment=attachment,
                )
                if url:
                    body = body.replace(
                        attachment.url,
                        url,
                    )
                else:
                    substitution = f'(Attachment "{attachment_name}" not available due to upload error)'
                    body = self._substitute_attachment(
                        body=body,
                        attachment=attachment,
                        substitution=substitution,
                    )
            else:
                substitution = f'(Attachment "{attachment_name}" not available due to export script’s configuration)'
                body = self._substitute_attachment(
                    body=body,
                    attachment=attachment,
                    substitution=substitution,
                )
        return body

    def _extract_attachment_name(
        self,
        body: str,
        attachment: Attachment,
    ) -> str:
        attachment_name_regex = self._attachment_name_regex_template.substitute(
            url=re.escape(attachment.url),
        )
        matches = re.findall(
            attachment_name_regex,
            body,
        )
        if matches:
            return cast(str, matches[0])
        return attachment.original_name

    def _substitute_attachment(
        self,
        body: str,
        attachment: Attachment,
        substitution: str,
    ) -> str:
        attachment_substitute_regex = self._attachment_substitute_regex_template.substitute(
            url=re.escape(attachment.url),
        )
        return re.sub(
            attachment_substitute_regex,
            substitution,
            body,
        )

    def _ensure_auth(
        self,
    ) -> None:
        try:
            user_id = self._github.get_user().id
        except GithubException as e:
            raise GitHubTrackerClientError('Unable to authenticate to GitHub') from e
        if not isinstance(user_id, int):
            raise GitHubTrackerClientError('Unable to authenticate to GitHub')



"""Models and functions used for testing of YesWeHack and tracker clients."""
from typing import Optional, cast

from ywh2bt.core.api.tracker import TrackerClientError
from ywh2bt.core.api.yeswehack import YesWeHackApiClientError
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.configuration.tracker import TrackerConfiguration, Trackers
from ywh2bt.core.configuration.yeswehack import YesWeHackConfiguration, YesWeHackConfigurations
from ywh2bt.core.mixins.tracker_clients import TrackerClientsMixin
from ywh2bt.core.mixins.yeswehack_api_clients import YesWeHackApiClientsMixin
from ywh2bt.core.tester.error import TesterError
from ywh2bt.core.tester.listener import (
    NoOpTesterListener,
    TesterEndEvent,
    TesterEndTrackerEvent,
    TesterEndYesWeHackEvent,
    TesterListener,
    TesterStartEvent,
    TesterStartTrackerEvent,
    TesterStartYesWeHackEvent,
)


class Tester(YesWeHackApiClientsMixin, TrackerClientsMixin):
    """A class used for testing of YesWeHack and tracker clients."""

    _configuration: RootConfiguration
    _listener: TesterListener

    def __init__(
        self,
        configuration: RootConfiguration,
        listener: Optional[TesterListener] = None,
    ):
        """
        Initialize self.

        Args:
            configuration: a configuration
            listener: an observer that will receive test events
        """
        YesWeHackApiClientsMixin.__init__(self)  # noqa: WPS609  # type: ignore
        TrackerClientsMixin.__init__(self)  # noqa: WPS609  # type: ignore
        self._configuration = configuration
        self._listener = listener or NoOpTesterListener()

    def test(
        self,
    ) -> None:
        """Test YesWeHack and tracker clients."""
        self._listener.on_event(
            event=TesterStartEvent(
                configuration=self._configuration,
            ),
        )
        self._test_yeswehack()
        self._test_bugtrackers()
        self._listener.on_event(
            event=TesterEndEvent(
                configuration=self._configuration,
            ),
        )

    def _test_yeswehack(
        self,
    ) -> None:
        yeswehack_configurations = cast(YesWeHackConfigurations, self._configuration.yeswehack)
        for yeswehack_name, yeswehack_configuration in yeswehack_configurations.items():
            self._test_yeswehack_configuration(
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
            )

    def _test_yeswehack_configuration(
        self,
        yeswehack_name: str,
        yeswehack_configuration: YesWeHackConfiguration,
    ) -> None:
        self._listener.on_event(
            event=TesterStartYesWeHackEvent(
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
            ),
        )
        yeswehack_client = self.get_yeswehack_api_client(
            configuration=yeswehack_configuration,
        )
        try:
            yeswehack_client.test()
        except YesWeHackApiClientError as e:
            raise TesterError(f'Test for yeswehack {yeswehack_name} has failed') from e
        self._listener.on_event(
            event=TesterEndYesWeHackEvent(
                yeswehack_name=yeswehack_name,
                yeswehack_configuration=yeswehack_configuration,
            ),
        )

    def _test_bugtrackers(
        self,
    ) -> None:
        bugtracker_configurations = cast(Trackers, self._configuration.bugtrackers)
        for bugtracker_name, bugtracker_configuration in bugtracker_configurations.items():
            self._test_bugtracker(
                bugtracker_name=bugtracker_name,
                bugtracker_configuration=bugtracker_configuration,
            )

    def _test_bugtracker(
        self,
        bugtracker_name: str,
        bugtracker_configuration: TrackerConfiguration,
    ) -> None:
        self._listener.on_event(
            event=TesterStartTrackerEvent(
                tracker_name=bugtracker_name,
                tracker_configuration=bugtracker_configuration,
            ),
        )
        tracker_client = self.get_tracker_client(
            configuration=bugtracker_configuration,
        )
        try:
            tracker_client.test()
        except TrackerClientError as e:
            raise TesterError(f'Test for bugtracker {bugtracker_name} has failed') from e
        self._listener.on_event(
            event=TesterEndTrackerEvent(
                tracker_name=bugtracker_name,
                tracker_configuration=bugtracker_configuration,
            ),
        )



"""Models and functions related to GUI tester thread."""
from datetime import datetime
from typing import cast

from PySide2.QtCore import QObject, QThread, Signal, SignalInstance
from singledispatchmethod import singledispatchmethod

from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.error import error_to_string
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.tester.listener import (
    TesterEndEvent,
    TesterEndTrackerEvent,
    TesterEndYesWeHackEvent,
    TesterEvent,
    TesterListener,
    TesterStartEvent,
    TesterStartTrackerEvent,
    TesterStartYesWeHackEvent,
)
from ywh2bt.core.tester.tester import Tester
from ywh2bt.gui.widgets.logs_widget import LogEntry, LogType
from ywh2bt.gui.widgets.root_configuration_entry import RootConfigurationEntry
from ywh2bt.gui.widgets.typing import as_signal_instance


class TesterThread(QThread):
    """A thread for testing configuration."""

    test_started: Signal = Signal()
    test_ended: Signal = Signal()

    _entry: RootConfigurationEntry
    _log_entry_available_signal: SignalInstance
    _tester: Tester

    def __init__(
        self,
        parent: QObject,
        entry: RootConfigurationEntry,
        log_entry_available_signal: SignalInstance,
    ):
        """
        Initialize self.

        Args:
            parent: a parent widget
            entry: a root configuration entry
            log_entry_available_signal: a signal that receive events from the thread
        """
        super().__init__(
            parent,
        )
        self._entry = entry
        self._log_entry_available_signal = log_entry_available_signal
        observer = _TesterListener(
            entry=entry,
            log_entry_available_signal=log_entry_available_signal,
        )
        self._tester = Tester(
            configuration=cast(RootConfiguration, entry.configuration),
            listener=observer,
        )

    def run(
        self,
    ) -> None:
        """Run the thread."""
        as_signal_instance(self.test_started).emit()
        try:
            self._tester.test()
        except CoreException as e:
            self._log_entry_available_signal.emit(
                LogEntry(
                    date_time=datetime.now(),
                    context=f'{self._entry.name}',
                    message=error_to_string(
                        error=e,
                    ),
                    log_type=LogType.error,
                ),
            )
        as_signal_instance(self.test_ended).emit()


class _TesterListener(TesterListener):
    _entry: RootConfigurationEntry
    _log_entry_available_signal: SignalInstance

    def __init__(
        self,
        entry: RootConfigurationEntry,
        log_entry_available_signal: SignalInstance,
    ):
        self._entry = entry
        self._log_entry_available_signal = log_entry_available_signal

    def _log_message(
        self,
        message: str,
        log_type: LogType = LogType.standard,
    ) -> None:
        self._log_entry_available_signal.emit(
            LogEntry(
                date_time=datetime.now(),
                context=f'{self._entry.name}',
                message=message,
                log_type=log_type,
            ),
        )

    def on_event(
        self,
        event: TesterEvent,
    ) -> None:
        self._on_event(event)

    @singledispatchmethod
    def _on_event(
        self,
        event: TesterEvent,
    ) -> None:
        self._log_message(
            message=f'Unhandled event {event}',
        )

    @_on_event.register
    def _on_start(
        self,
        event: TesterStartEvent,
    ) -> None:
        self._log_message(
            message='Starting test',
        )

    @_on_event.register
    def _on_end(
        self,
        event: TesterEndEvent,
    ) -> None:
        self._log_message(
            log_type=LogType.success,
            message='Test done',
        )

    @_on_event.register
    def _on_start_test_yeswehack(
        self,
        event: TesterStartYesWeHackEvent,
    ) -> None:
        self._log_message(
            message=f'Testing YesWeHack {event.yeswehack_name}',
        )

    @_on_event.register
    def _on_end_test_yeswehack(
        self,
        event: TesterEndYesWeHackEvent,
    ) -> None:
        self._log_message(
            log_type=LogType.success,
            message=f'YesWeHack {event.yeswehack_name}: OK',
        )

    @_on_event.register
    def _on_start_test_tracker(
        self,
        event: TesterStartTrackerEvent,
    ) -> None:
        self._log_message(
            message=f'Testing tracker {event.tracker_name}',
        )

    @_on_event.register
    def _on_end_test_tracker(
        self,
        event: TesterEndTrackerEvent,
    ) -> None:
        self._log_message(
            log_type=LogType.success,
            message=f'Tracker {event.tracker_name}: OK',
        )



"""Models and functions used for data synchronisation between YesWeHack and Jira trackers."""
from typing import Any, List, Optional, Tuple, cast
from urllib.parse import quote

from jira import Comment as JIRAComment, Issue as JIRAIssue, JIRA, JIRAError  # type: ignore
from jira.utils import CaseInsensitiveDict  # type: ignore

from ywh2bt.core.api.models.report import Attachment, Log, Report
from ywh2bt.core.api.tracker import TrackerClient, TrackerClientError, TrackerComment, TrackerComments, TrackerIssue
from ywh2bt.core.api.trackers.jira.formatter import JiraReportMessageFormatter
from ywh2bt.core.configuration.trackers.jira import JiraConfiguration


# hot patching shenanigans
def _case_insensitive_dict_init_hot_patch(
    obj: CaseInsensitiveDict,
    *args: Any,
    **kw: Any,
) -> None:
    super(CaseInsensitiveDict, obj).__init__(*args, *kw)  # noqa: WPS608, WPS613
    obj_copy = super(CaseInsensitiveDict, obj).copy()  # noqa: WPS608, WPS613
    obj.itemlist = {}
    for key, value in obj_copy.items():
        if key != key.lower():
            obj[key.lower()] = value
            obj.pop(key, None)


CaseInsensitiveDict.__init__ = _case_insensitive_dict_init_hot_patch  # noqa: WPS609


class JiraTrackerClientError(TrackerClientError):
    """A Jira tracker client error."""


class JiraTrackerClient(TrackerClient[JiraConfiguration]):
    """A Jira tracker client."""

    _jira: Optional[JIRA]
    _message_formatter: JiraReportMessageFormatter

    def __init__(
        self,
        configuration: JiraConfiguration,
    ):
        """
        Initialize self.

        Args:
            configuration: a Jira configuration
        """
        super().__init__(
            configuration=configuration,
        )
        self._jira = None
        self._message_formatter = JiraReportMessageFormatter()

    @property
    def tracker_type(self) -> str:
        """
        Get the type of the tracker client.

        Returns:
            the type of the  tracker client
        """
        return 'Jira'

    def _build_tracker_issue(
        self,
        issue_id: str,
        issue_url: str,
        is_closed: bool,
    ) -> TrackerIssue:
        return TrackerIssue(
            tracker_url=cast(str, self.configuration.url),
            project=cast(str, self.configuration.project),
            issue_id=issue_id,
            issue_url=issue_url,
            is_closed=is_closed,
        )

    def get_tracker_issue(
        self,
        issue_id: str,
    ) -> Optional[TrackerIssue]:
        """
        Get a tracker issue.

        Args:
            issue_id: an issue id

        Returns:
            The issue if it exists, else None
        """
        try:
            jira_issue = self._get_issue(
                issue_id=issue_id,
            )
        except JiraTrackerClientError:
            return None
        return self._build_tracker_issue(
            issue_id=issue_id,
            issue_url=jira_issue.permalink(),
            is_closed=all((
                jira_issue.fields.status is not None,
                jira_issue.fields.status.name == self.configuration.issue_closed_status,
            )),
        )

    def send_report(
        self,
        report: Report,
    ) -> TrackerIssue:
        """
        Send a report to the tracker.

        Args:
            report: a report

        Returns:
            information about the sent report
        """
        self._ensure_auth()
        title = self._message_formatter.format_report_title(
            report=report,
        )
        description = self._message_formatter.format_report_description(
            report=report,
        )
        jira_issue = self._create_issue(
            title=title,
            description=description,
        )
        uploads = self._upload_attachments(
            issue=jira_issue,
            attachments=report.attachments,
        )
        for attachment_url, uploaded_url in uploads:
            description = description.replace(
                attachment_url,
                uploaded_url,
            )
        jira_issue.update(
            description=description,
        )
        return self._build_tracker_issue(
            issue_id=jira_issue.key,
            issue_url=jira_issue.permalink(),
            is_closed=False,
        )

    def send_logs(
        self,
        tracker_issue: TrackerIssue,
        logs: List[Log],
    ) -> TrackerComments:
        """
        Send logs to the tracker.

        Args:
            tracker_issue: information about the tracker issue
            logs: a list of logs

        Returns:
            information about the sent logs
        """
        jira_issue = self._get_issue(
            issue_id=tracker_issue.issue_id,
        )
        tracker_comments = TrackerComments(
            tracker_issue=tracker_issue,
            added_comments=[],
        )
        for log in logs:
            jira_comment = self._add_comment(
                issue=jira_issue,
                log=log,
            )
            tracker_comments.added_comments.append(
                TrackerComment(
                    comment_id=jira_comment.id,
                ),
            )
        return tracker_comments

    def test(
        self,
    ) -> None:
        """Test the client."""
        self._ensure_auth()

    def _get_issue(
        self,
        issue_id: str,
    ) -> JIRAIssue:
        try:
            return self._get_client().issue(
                id=issue_id,
            )
        except JIRAError as e:
            raise JiraTrackerClientError(
                f'Unable to get JIRA issue {issue_id} in project {self.configuration.project}',
            ) from e

    def _create_issue(
        self,
        title: str,
        description: str,
    ) -> JIRAIssue:
        fields = {
            'project': {
                'key': self.configuration.project,
            },
            'summary': title,
            'description': description,
            'issuetype': {
                'name': self.configuration.issuetype,
            },
        }
        try:
            return self._get_client().create_issue(
                fields=fields,
            )
        except JIRAError as e:
            raise JiraTrackerClientError(
                f'Unable to create JIRA issue for project {self.configuration.project}',
            ) from e

    def _add_comment(
        self,
        issue: JIRAIssue,
        log: Log,
    ) -> JIRAComment:
        comment_body = self._message_formatter.format_log(
            log=log,
        )
        uploads = self._upload_attachments(
            issue=issue,
            attachments=log.attachments,
        )
        for attachment_url, uploaded_url in uploads:
            comment_body = comment_body.replace(
                attachment_url,
                uploaded_url,
            )
        try:
            return self._get_client().add_comment(
                issue=str(issue),
                body=comment_body,
            )
        except JIRAError as e:
            raise JiraTrackerClientError(
                f'Unable to add JIRA comment for issue {issue} in project {self.configuration.project}',
            ) from e

    def _upload_attachments(
        self,
        issue: JIRAIssue,
        attachments: List[Attachment],
    ) -> List[Tuple[str, str]]:
        uploads = []
        for attachment in attachments:
            try:
                issue_attachment = self._get_client().add_attachment(
                    issue=issue.key,
                    filename=attachment.original_name,
                    attachment=attachment.data,
                )
            except JIRAError as e:
                raise JiraTrackerClientError(
                    f'Unable to upload attachments for project {self.configuration.project} to JIRA',
                ) from e
            issue_attachment_url = quote(
                issue_attachment.content,
                safe=':/?&=',
            )
            uploads.append(
                (
                    attachment.url,
                    issue_attachment_url,
                ),
            )
        return uploads

    def _get_client(
        self,
    ) -> JIRA:
        if not self._jira:
            self._jira = self._build_client()
        return self._jira

    def _build_client(
        self,
    ) -> JIRA:
        configuration = self.configuration
        try:
            return JIRA(
                server=configuration.url,
                basic_auth=(
                    configuration.login,
                    configuration.password,
                ),
                options={
                    'verify': configuration.verify,
                },
            )
        except JIRAError as e:
            raise JiraTrackerClientError('Unable to connect to JIRA') from e

    def _ensure_auth(
        self,
    ) -> None:
        try:
            self._get_client().myself()
        except JIRAError as e:
            raise JiraTrackerClientError('Unable to authenticate to JIRA') from e



"""Models and functions related to GUI synchronizer thread."""
from datetime import datetime
from typing import cast

from PySide2.QtCore import QObject, QThread, Signal, SignalInstance
from singledispatchmethod import singledispatchmethod

from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.error import error_to_string
from ywh2bt.core.exceptions import CoreException
from ywh2bt.core.synchronizer.listener import (
    SynchronizerEndEvent,
    SynchronizerEndFetchReportsEvent,
    SynchronizerEndSendReportEvent,
    SynchronizerEvent,
    SynchronizerListener,
    SynchronizerStartEvent,
    SynchronizerStartFetchReportsEvent,
    SynchronizerStartSendReportEvent,
)
from ywh2bt.core.synchronizer.synchronizer import Synchronizer
from ywh2bt.gui.widgets.logs_widget import LogEntry, LogType
from ywh2bt.gui.widgets.root_configuration_entry import RootConfigurationEntry
from ywh2bt.gui.widgets.typing import as_signal_instance


class SynchronizerThread(QThread):
    """A thread for synchronization."""

    synchronization_started: Signal = Signal()
    synchronization_ended: Signal = Signal()

    _entry: RootConfigurationEntry
    _log_entry_available_signal: SignalInstance
    _synchronizer: Synchronizer

    def __init__(
        self,
        parent: QObject,
        entry: RootConfigurationEntry,
        log_entry_available_signal: SignalInstance,
    ):
        """
        Initialize self.

        Args:
            parent: a parent widget
            entry: a root configuration entry
            log_entry_available_signal: a signal that receive events from the thread
        """
        super().__init__(
            parent,
        )
        self._entry = entry
        self._log_entry_available_signal = log_entry_available_signal
        observer = _SynchronizerListener(
            entry=entry,
            log_entry_available_signal=log_entry_available_signal,
        )
        self._synchronizer = Synchronizer(
            configuration=cast(RootConfiguration, entry.configuration),
            listener=observer,
        )

    def run(
        self,
    ) -> None:
        """Run the thread."""
        as_signal_instance(self.synchronization_started).emit()
        try:
            self._synchronizer.synchronize()
        except CoreException as e:
            self._log_entry_available_signal.emit(
                LogEntry(
                    date_time=datetime.now(),
                    context=f'{self._entry.name}',
                    message=error_to_string(
                        error=e,
                    ),
                    log_type=LogType.error,
                ),
            )
        as_signal_instance(self.synchronization_ended).emit()


class _SynchronizerListener(SynchronizerListener):
    _entry: RootConfigurationEntry
    _log_entry_available_signal: SignalInstance

    def __init__(
        self,
        entry: RootConfigurationEntry,
        log_entry_available_signal: SignalInstance,
    ):
        self._entry = entry
        self._log_entry_available_signal = log_entry_available_signal

    def _log_message(
        self,
        message: str,
        log_type: LogType = LogType.standard,
    ) -> None:
        self._log_entry_available_signal.emit(
            LogEntry(
                date_time=datetime.now(),
                context=f'{self._entry.name}',
                message=message,
                log_type=log_type,
            ),
        )

    def on_event(
        self,
        event: SynchronizerEvent,
    ) -> None:
        self._on_event(event)

    @singledispatchmethod
    def _on_event(
        self,
        event: SynchronizerEvent,
    ) -> None:
        self._log_message(
            message=f'Unhandled event {event}',
        )

    @_on_event.register
    def _on_start(
        self,
        event: SynchronizerStartEvent,
    ) -> None:
        self._log_message(
            message='Starting synchronization...',
        )

    @_on_event.register
    def _on_end(
        self,
        event: SynchronizerEndEvent,
    ) -> None:
        self._log_message(
            log_type=LogType.success,
            message='Synchronization done.',
        )

    @_on_event.register
    def _on_start_fetch_reports(
        self,
        event: SynchronizerStartFetchReportsEvent,
    ) -> None:
        yeswehack_name = event.yeswehack_name
        program_slug = event.program_slug
        self._log_message(
            message=f'Processing YesWeHack "{yeswehack_name}"...',
        )
        self._log_message(
            message=f'Fetching reports for program "{program_slug}"...',
        )

    @_on_event.register
    def _on_end_fetch_reports(
        self,
        event: SynchronizerEndFetchReportsEvent,
    ) -> None:
        reports_count = len(event.reports)
        self._log_message(
            log_type=LogType.success,
            message=f'{reports_count} report(s) fetched.',
        )

    @_on_event.register
    def _on_start_send_report(
        self,
        event: SynchronizerStartSendReportEvent,
    ) -> None:
        report = event.report
        report_details = f'#{report.report_id} ({report.title})'
        self._log_message(
            message=f'Processing report {report_details} with "{event.tracker_name}"...',
        )

    @_on_event.register
    def _on_end_send_report(
        self,
        event: SynchronizerEndSendReportEvent,
    ) -> None:
        result = event.result
        if event.is_existing_issue:
            if result.added_comments:
                issue_action = 'updated'
            else:
                issue_action = 'untouched'
        else:
            issue_action = 'added'
        report_details = f'#{event.report.report_id} ({event.report.title})'
        comments_details = f'{len(result.added_comments)} comment(s) added'
        tracking_status_action = 'updated' if event.tracking_status_updated else 'unchanged'
        message = ' | '.join(
            (
                f'{result.tracker_issue.issue_url} ({issue_action} ; {comments_details})',
                f'tracking status {tracking_status_action}',
            ),
        )
        self._log_message(
            log_type=LogType.success,
            message=f'Processed report {report_details} with "{event.tracker_name}": {message}.',
        )



"""Models and functions used for observing synchronisation between YesWeHack and trackers."""
from abc import ABC
from dataclasses import dataclass
from typing import List

from ywh2bt.core.api.models.report import Report
from ywh2bt.core.api.tracker import TrackerComments
from ywh2bt.core.configuration.root import RootConfiguration
from ywh2bt.core.configuration.yeswehack import YesWeHackConfiguration
from ywh2bt.core.listener import Event, Listener


@dataclass
class SynchronizerEvent(Event, ABC):
    """A base event sent by a Synchronizer."""


@dataclass
class SynchronizerStartEvent(SynchronizerEvent):
    """An event sent when a synchronization starts."""

    configuration: RootConfiguration


@dataclass
class SynchronizerEndEvent(SynchronizerEvent):
    """An event sent when a synchronization ends."""

    configuration: RootConfiguration


@dataclass
class SynchronizerStartFetchReportsEvent(SynchronizerEvent):
    """An event sent when fetching of reports starts."""

    configuration: RootConfiguration
    yeswehack_name: str
    yeswehack_configuration: YesWeHackConfiguration
    program_slug: str


@dataclass
class SynchronizerEndFetchReportsEvent(SynchronizerEvent):
    """An event sent when fetching of reports ends."""

    configuration: RootConfiguration
    yeswehack_name: str
    yeswehack_configuration: YesWeHackConfiguration
    program_slug: str
    reports: List[Report]


@dataclass
class SynchronizerStartSendReportEvent(SynchronizerEvent):
    """An event sent when a synchronization of a report starts."""

    configuration: RootConfiguration
    yeswehack_name: str
    yeswehack_configuration: YesWeHackConfiguration
    program_slug: str
    tracker_name: str
    report: Report


@dataclass
class SynchronizerEndSendReportEvent(SynchronizerEvent):
    """An event sent when a synchronization of a report ends."""

    configuration: RootConfiguration
    yeswehack_name: str
    yeswehack_configuration: YesWeHackConfiguration
    program_slug: str
    tracker_name: str
    report: Report
    is_existing_issue: bool
    tracking_status_updated: bool
    result: TrackerComments


class SynchronizerListener(Listener[SynchronizerEvent], ABC):
    """A listener receiving events from a Synchronizer."""


class NoOpSynchronizerListener(SynchronizerListener):
    """A listener receiving events from a Synchronizer and doing nothing."""

    def on_event(
        self,
        event: SynchronizerEvent,
    ) -> None:
        """
        Receive an event and do nothing.

        Args:
            event: an event
        """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ywh2bt',
 'ywh2bt.cli',
 'ywh2bt.cli.commands',
 'ywh2bt.core',
 'ywh2bt.core.api',
 'ywh2bt.core.api.formatter',
 'ywh2bt.core.api.models',
 'ywh2bt.core.api.trackers',
 'ywh2bt.core.api.trackers.github',
 'ywh2bt.core.api.trackers.jira',
 'ywh2bt.core.configuration',
 'ywh2bt.core.configuration.trackers',
 'ywh2bt.core.crypt',
 'ywh2bt.core.mixins',
 'ywh2bt.core.schema',
 'ywh2bt.core.serializers',
 'ywh2bt.core.state',
 'ywh2bt.core.synchronizer',
 'ywh2bt.core.tester',
 'ywh2bt.gui',
 'ywh2bt.gui.dialog',
 'ywh2bt.gui.widgets',
 'ywh2bt.gui.widgets.attribute',
 'ywh2bt.gui.widgets.thread']

package_data = \
{'': ['*'],
 'ywh2bt.gui': ['resources/icons/*',
                'resources/icons/types/*',
                'resources/icons/types/TrackerConfiguration/*',
                'resources/translations/*']}

install_requires = \
['PyGithub>=1.53,<2.0',
 'PySide2>=5.15.1,<6.0.0',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'html2text>=2020.1.16,<2021.0.0',
 'jira>=3.0a2,<4.0',
 'lxml>=4.5.2,<5.0.0',
 'markdown>=3.3.3,<4.0.0',
 'python-gitlab>=2.5.0,<3.0.0',
 'requests-toolbelt>=0.9.1,<0.10.0',
 'requests>=2.24.0,<3.0.0',
 'ruamel.yaml>=0.16.12,<0.17.0',
 'singledispatchmethod>=1.0,<2.0',
 'tomlkit>=0.7.0,<0.8.0',
 'typing-extensions',
 'yeswehack<1.0.0']

entry_points = \
{'console_scripts': ['ywh2bt = ywh2bt.cli.main:run',
                     'ywh2bt-gui = ywh2bt.gui.main:run']}

setup_kwargs = {
    'name': 'ywh2bt',
    'version': '2.0.1',
    'description': 'ywh2bt - YesWeHack to Bug Tracker',
    'long_description': '# ywh2bt\n\nywh2bt is a tool to integrate your bug tracking system(s) with [YesWeHack platform][YesWeHack-Platform].\nIt automatically creates issues in your bug tracking system for all your program\'s report,\nand add to the concerned reports the link to the issue.\n\nThis tool requires you to have "Use Apps API" right on [YesWeHack platform][YesWeHack-Platform],\nand a custom HTTP header value to put in your configuration. \nBoth of them can be obtained by e-mailing us at support@yeswehack.com.\n\n- [Features](#features)\n    - [Supported trackers](#supported-trackers)\n- [Requirements](#requirements)\n- [Installation](#installation)\n- [GUI](#gui)\n    - [Usage](#usage)\n    - [Screenshots](#screenshots)\n- [Command line](#command-line)\n    - [`ywh2bt`](#ywh2bt-1)\n        - [Commands](#commands)\n        - [Example usages](#example-usages)\n- [Supported configuration file formats](#supported-configuration-file-formats)\n- [Known limitations and specific behaviours](#known-limitations-and-specific-behaviours)\n- [Changelog](#changelog)\n- [Local development](#local-development)\n    - [Requirements](#requirements-1)\n    - [Installation](#installation-1)\n    - [Usage](#usage-1)\n\n## Features\n\n- synchronization from [YesWeHack platform][YesWeHack-Platform] to trackers:\n    - platform reports to tracker issues\n    - reports logs/comments to issues comments\n- creation, modification, synchronization, validation, conversion of configuration files through a GUI\n- validation of configuration files\n- format conversion of configuration files\n\n### Supported trackers\n\n- github\n- gitlab\n- jira / jiracloud\n\n## Requirements\n\n- `python` >= 3.7,<=3.9\n- [`pip`](https://pip.pypa.io/en/stable/installing/)\n\n## Installation\n\n```sh\npip install ywh2bt\n```\n\n## GUI\n\nThe GUI provides assistance to create, modify and validate/test configurations. \nIt also allows synchronization with bug trackers.\n\nTo run it, simply type `ywh2bt-gui` in a shell.\n\n### Usage\n\n- Changes to the configuration can be made either in the configuration tab or in the "Raw" tab ; \n  changes made in one tab are automatically reflected in the other tab.\n- Hovering labels and buttons with the mouse pointer often reveals more information in a floating tooltip \n  or in the status bar.\n- A description of the schema of the configuration files is accessible via the "Help > Schema documentation" menu\n  or by clicking on the ![Icon for help button](doc/img/icon-help.png) button in the main toolbar.\n\n### Screenshots\n\n- [example.yml](doc/examples/example.yml) configuration:\n\n![Screenshot of GUI with loaded example file](doc/img/screenshot-gui-example.png)\n\n- [empty configuration](doc/img/screenshot-gui-new.png)\n\n## Command line\n\n### `ywh2bt`\n\nMain script used to execute synchronization, validate and test configurations.\n\nUsage: `ywh2bt [command]`. See `ywh2bt -h` or `ywh2bt [command] -h` for detailed help.\n\n#### Commands\n\n- `validate`: validate a configuration file (mandatory fields, data types, ...)\n- `test`: test the connection to the trackers\n- `convert`: convert a configuration file into another format\n- `synchronize` (alias `sync`): synchronize trackers with YesWeHack reports\n- `schema`: dump a schema of the structure of the configuration files in [Json-Schema][Json-Schema], markdown \n  or plaintext\n\n#### Example usages\n\nValidation:\n```sh\n$ ywh2bt validate \\\n    --config-file=my-config.yml \\\n    --config-format=yaml && echo OK\nOK\n```\n\nConversion (`yaml` to `json`):\n```sh\n$ ywh2bt convert \\\n    --config-file=my-config.yml \\\n    --config-format=yaml \\\n    --destination-file=/tmp/cfg.json \\\n    --destination-format=json\n```\n\nSynchronization:\n```sh\n$ ywh2bt synchronize --config-file=my-config.json --config-format=json\n[2020-12-21 10:20:58.881315] Starting synchronization:\n[2020-12-21 10:20:58.881608]   Processing YesWeHack "yeswehack1": \n[2020-12-21 10:20:58.881627]     Fetching reports for program "my-program": 2 report(s)\n[2020-12-21 10:21:08.341460]     Processing report #123 (CVE-2017-11882 on program) with "my-github": https://github.com/user/project/issues/420 (untouched ; 0 comment(s) added) | tracking status unchanged\n[2020-12-21 10:21:09.656178]     Processing report #96 (I found a bug) with "my-github": https://github.com/user/project/issues/987 (created ; 3 comment(s) added) | tracking status updated\n[2020-12-21 10:21:10.773688] Synchronization done.\n```\n\n## Supported configuration file formats\n\n- `yaml` (legacy)\n- `json`\n\nUse `ywh2bt schema -f json` to obtain a [Json-Schema][Json-Schema] describing the format.\nBoth `yaml` and `json` configuration files should conform to the schema. \n\n## Known limitations and specific behaviours\n\n- Apps API doesn\'t require TOTP authentication, even if corresponding user has TOTP enabled.  \n  However, on a secured program, information is limited for user with TOTP disabled, even in apps.  \n  As a consequence, to allow proper bug tracking integration on a secured program,\n  program consumer must have TOTP enabled and, in BTI configuration TOTP must be set to `false`.\n- References to a same uploaded attachment in different comments is not supported yet,\n  i.e., if an attachment is referenced (either displayed inline or as a link) in several comments,\n  only first one will be correctly handled.\n- Manually tracked reports (i.e., where a manager directly set the Tracking status to "tracked") \n  are also integrated in the tracker the way they are when a manager set "Ask for integration".\n- Since v2.0.0, unlike in previous versions, setting a tracked report back to "Ask for integration" \n  won\'t create a new issue in the tracker but update the existing one.\n\n## Changelog\n\n- v0.* to v2.0.0:\n    - behavior changes:\n        - reports logs can selectively be synchronized with the trackers:\n            - public comments\n            - private comments\n            - report details changes\n            - report status changes\n            - rewards\n        - a program can now only be synchronized with 1 tracker\n    - added support for JSON configuration files\n    - removed `ywh-bugtracker` command (use `ywh2bt synchronize`)\n    - added `ywh2bt` command:\n        - added `ywh2bt synchronize`:\n            - note: `ywh2bt synchronize --config-file FILE --config-format FORMAT` \n              is the equivalent of `ywh-bugtracker -n -f FILE` in v0.*\n        - added `ywh2bt validate`\n        - added `ywh2bt test`\n        - added `ywh2bt convert`\n        - added `ywh2bt schema`\n    - removed command line interactive mode\n    - added GUI via `ywh2bt-gui` command\n\n## Local development\n\n### Requirements\n\n- [`poetry`](https://python-poetry.org/) (`pip install poetry`)\n\n### Installation\n\n- `make install` (or `poetry install`): creates a virtualenv and install dependencies\n\n### Usage\n\nInstead of `ywh2bt [command]`, run commands using `poetry run ywh2bt [command]`.\n\nSame goes for `ywh2bt-gui`, run `poetry run ywh2bt-gui` instead.\n\n\n[YesWeHack-Platform]: https://www.yeswehack.com/\n[Json-Schema]: https://json-schema.org/specification.html',
    'author': 'm.honel',
    'author_email': 'm.honel@yeswehack.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yeswehack/ywh2bugtracker',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<3.10',
}


setup(**setup_kwargs)



"""Synchronizer mapping between tracker configurations and tracker clients."""
from types import MappingProxyType
from typing import Any, Mapping, Type

from ywh2bt.core.api.tracker import TrackerClient
from ywh2bt.core.api.trackers.github.tracker import GitHubTrackerClient
from ywh2bt.core.api.trackers.gitlab import GitLabTrackerClient
from ywh2bt.core.api.trackers.jira.tracker import JiraTrackerClient
from ywh2bt.core.configuration.tracker import TrackerConfiguration
from ywh2bt.core.configuration.trackers.github import GitHubConfiguration
from ywh2bt.core.configuration.trackers.gitlab import GitLabConfiguration
from ywh2bt.core.configuration.trackers.jira import JiraConfiguration

TrackerClientClassesType = Mapping[
    Type[TrackerConfiguration],
    Type[TrackerClient[Any]],
]
TRACKER_CLIENT_CLASSES: TrackerClientClassesType = MappingProxyType({
    GitHubConfiguration: GitHubTrackerClient,
    GitLabConfiguration: GitLabTrackerClient,
    JiraConfiguration: JiraTrackerClient,
})
